.table{min-width:40%;margin:1.5625rem 0;border-collapse:collapse}.table thead tr{background-color:var(--contrast);color:var(--text-color);text-align:left}.table th,.table td{padding:.75rem .94rem;-webkit-hyphens:manual;hyphens:manual;word-break:break-all}.table tbody tr{border-bottom:thin solid var(--contrast-2)}.table tbody tr:nth-child(2n){background-color:var(--secondary-bg)}.table tbody tr:last-child{border-bottom:medium solid var(--contrast-2)}
/*# sourceMappingURL=table.css.map */
