{% extends "../../base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/form.css')}}">
{% end %}

{% block body %}
    <form method="post" action="{{fix_url('/zitate/create-wrong-quote')}}">
        <label>
            Zitat:
             <select name="quote-2">
                 {% for quote in quotes %}
                     {% if isinstance(quote, str) %}
                         <option value="{{quote.strip()}}">{{quote.strip()}}</option>
                     {% else %}
                         <option value="{{quote.quote.strip()}}">{{quote}}</option>
                     {% end %}
                 {% end %}
             </select>
        </label>

        <label>
            Richtiger Autor:
            <select name="real-author-2">
                {% for author in real_authors %}
                    {% if isinstance(author, str) %}
                        <option value="{{author.strip()}}">{{author.strip()}}</option>
                    {% else %}
                        <option value="{{author.name.strip()}}">{{author}}</option>
                    {% end %}
                {% end %}
            </select>
        </label>

        <label>
            Falscher Autor:
            <select name="fake-author-2">
                {% for author in fake_authors %}
                    {% if isinstance(author, str) %}
                        <option value="{{author.strip()}}">{{author.strip()}}</option>
                    {% else %}
                        <option value="{{author.name.strip()}}">{{author}}</option>
                    {% end %}
                {% end %}
            </select>
        </label>

        <input type="text" class="hidden" name="nicht-erstellen" value="">

        <input id="submit-button"
               class="button"
               type="submit"
               value="Zitat erstellen">
    </form>
{% end %}
