:root{--main-light: hsl(0deg 100% 33%);--main-2-light: hsl(5deg 90% 22%);--text-color-light: hsl(0deg 0% 0%);--secondary-text-color-light: hsl(0deg 0% 15%);--bg-light: fixed radial-gradient( ellipse at bottom, hsl(180deg 30% 80%) 0%, hsl(180deg 30% 90%) 100% );--secondary-bg-light: hsl(180deg 10% 75%);--tertiary-bg-light: hsl(180deg 10% 60%);--contrast-light: hsl(195deg 100% 26%);--contrast-2-light: hsl(240deg 100% 50%);--main-dark: hsl(0deg 90% 48%);--main-2-dark: hsl(0deg 90% 62%);--text-color-dark: hsl(0deg 0% 90%);--secondary-text-color-dark: hsl(0deg 0% 70%);--bg-dark: fixed radial-gradient(ellipse at bottom, #1b2735 0%, #090a0f 100%);--secondary-bg-dark: hsl(0deg 0% 9%);--tertiary-bg-dark: hsl(0deg 0% 18%);--contrast-dark: hsl(240deg 100% 50%);--contrast-2-dark: hsl(195deg 100% 50%)}#snow p{position:fixed;width:10px;height:10px;border-radius:50%;margin:0;background:linear-gradient(#fff,#fff);filter:drop-shadow(0 0 10px var(--text-color))}main{text-shadow:0 0 10px var(--text-color)}header,#site-pane{box-shadow:0 0 10px var(--text-color)}textarea,select,input[type=email],input[type=range],input[type=search],input[type=text],input[type=number],[tooltip]:after{background:var(--secondary-bg)!important;background:light-dark(hsl(180deg 30% 90%),hsl(0deg 0% 0%))!important}
/*# sourceMappingURL=christmas.css.map */
