:root{--main-light: hsl(30deg 100% 60%);--main-2-light: hsl(33deg 100% 40%);--text-color-light: hsl(0deg 0% 0%);--secondary-text-color-light: hsl(0deg 0% 15%);--bg-light: hsl(180deg 19% 92%);--secondary-bg-light: hsl(180deg 9% 81%);--tertiary-bg-light: hsl(225deg 7% 77%);--contrast-light: hsl(195deg 100% 26%);--contrast-2-light: hsl(240deg 100% 50%);--main-dark: hsl(30deg 100% 60%);--main-2-dark: hsl(33deg 100% 40%);--text-color-dark: hsl(0deg 0% 90%);--secondary-text-color-dark: hsl(0deg 0% 70%);--bg-dark: hsl(0deg 0% 0%);--secondary-bg-dark: hsl(0deg 0% 9%);--tertiary-bg-dark: hsl(0deg 0% 18%);--contrast-dark: hsl(0deg 90% 48%);--contrast-2-dark: hsl(0deg 90% 62%)}
/*# sourceMappingURL=orange.css.map */
