.quote-a{color:var(--secondary-text-color)!important}h1:hover .header-id-link:before,h1:focus-within .header-id-link:before,h2:hover .header-id-link:before,h2:focus-within .header-id-link:before{content:"🔗"}#soundboard-rss{position:fixed;z-index:0;top:calc(var(--header-height) + .9rem);right:.9rem}#soundboard-rss img{width:1.7rem;height:1.7rem}input{margin-bottom:.3125rem}.marked{display:inline!important;color:var(--contrast-2)}
/*# sourceMappingURL=soundboard.css.map */
