@font-face{font-display:swap;font-family:comic;font-style:normal;src:local("Comic Sans MS"),url(/static/fonts/comic.woff2?v=0) format("woff2")}:root,[tooltip]:after{font-family:comic,serif}.openmoji{font-family:OpenMojiColor,comic,serif}@font-face{font-display:swap;font-family:comic-mono;font-style:normal;src:url(/static/fonts/comicmono.woff2?v=0) format("woff2")}pre *,.pre,.mono-font{font-family:comic-mono,ui-monospace,monospace}
/*# sourceMappingURL=comic.css.map */
