<header data-elastic-exclude id="header">
    <section id="title-container">
        <a aria-hidden="true"
           tabindex="-1"
           href="{{fix_url('/')}}"
           class="no-style-a-emoji"
           id="back-to-home">
            <img src="{{fix_static('img/logo.svg')}}" alt="Zurück zur Startseite">
        </a>
        <h1 id="title" short_title="{{short_title}}">{{title}}</h1>
    </section>
    {# NETCUP won't be shown if SHOW_HAMBURGER_MENU is False (that looks ugly) #}
    {% if settings and settings.get("SHOW_HAMBURGER_MENU") %}
    <div data-nosnippet>
        <img id="open-pane"
             src="{{fix_static('img/hamburger-menu-icon.svg')}}"
             alt="Menü">
        <nav id="site-pane" tabindex="-1">
            <h2 data-elastic-exclude>Unterseiten:</h2>
            <ul class="list">
                {% for i, info in enumerate(settings.get("MODULE_INFOS")) %}
                    {% if not info.hidden and info.path %}
                        <li>
                            <a href="{{fix_url(info.path)}}"
                               tooltip="{{info.description}}"
                               class="site-pane-a"
                               {% if i < 6 %}
                                   tooltip-position="bottom"
                               {% end %}
                               {% if info.path == "/js-lizenzen/" %}
                                   data-jslicense="1"
                               {% end %}
                            >{{info.name}}</a>
                            {% if info.sub_pages %}
                                <ul class="sub-list">
                                    {% for page_info in info.sub_pages %}
                                        {% if not page_info.hidden and page_info.path %}
                                            <li>
                                                <a href="{{fix_url(page_info.path)}}"
                                                   tooltip="{{page_info.description}}"
                                                   class="site-pane-a"
                                                >{{page_info.name}}</a>
                                            </li>
                                        {% end %}
                                    {% end %}
                                </ul>
                            {% end %}
                        </li>
                    {% end %}
                {% end %}
            </ul>
            {% if settings.get("CONTACT_USE_FORM") %}
                <a href="{{fix_url('/kontakt')}}"
                   rel="nofollow"
                   class="contact-anchor">KONTAKT</a>
            {% end %}
            {% if settings.get("NETCUP") %}
                <a id="oekostrom" href="{{fix_url('https://www.netcup.de/?ref=95236')}}">
                    <picture>
                        <source srcset="{{fix_static('img/netcup-oekostrom2.jxl')}}" type="image/jxl">
                        <img src="{{fix_static('img/netcup-oekostrom2.png')}}" alt="netcup Ökostrom">
                    </picture>
                </a>
            {% end %}
        </nav>
    </div>
    {% end %}
</header>
