#connection-state{display:inline-block;width:1.375em;height:1.375em;border-radius:50%;vertical-align:bottom}#connection-state[data-state=disconnected]{background-color:red}#connection-state[data-state=connected]{background-color:green}#connection-state[data-state=connecting]{background-color:orange}#current-user{display:inline}
/*# sourceMappingURL=emoji_chat.css.map */
