{% extends "../base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/form.css')}}">
{% end %}

{% block body %}
    <form>
        <label class="form-element">
            Das aktuelle Wort
            <input type="text"
                   name="word"
                   value="{{word}}">
        </label>

        <label class="form-element">
            Wörter, die vor diesem Wort kamen (Mit Komma getrennt)
            <input type="text"
                   name="before"
                   value="{{before}}">
        </label>

        {% raw form_appendix %}

        <input type="submit"
               id="submit"
               value="Eingabe">
    </form>

    <ul>
        {% for count, w in words %}
            <li>
                <a href="{{fix_url('/wordspiel-helfer', word=w, before=new_before)}}">{{w}}</a>
                ({{count}})
            </li>
        {% end %}
    </ul>
{% end %}
