@import"/static/css/comic.css?v=0";:root{--main-light: hsl(48deg 100% 50%);--main-2-light: hsl(30deg 100% 50%);--text-color-light: hsl(106deg 100% 50%);--secondary-text-color-light: hsl(159deg 100% 50%);--bg-light: hsl(303deg 100% 50%);--secondary-bg-light: hsl(289deg 100% 50%);--tertiary-bg-light: hsl(264deg 100% 50%);--contrast-light: hsl(207deg 100% 50%);--contrast-2-light: hsl(195deg 100% 50%);--main-dark: hsl(69deg 97% 40%);--main-2-dark: hsl(161deg 84% 27%);--text-color-dark: hsl(60deg 100% 58%);--secondary-text-color-dark: hsl(68deg 93% 36%);--bg-dark: hsl(285deg 70% 10%);--secondary-bg-dark: hsl(300deg 43% 6%);--tertiary-bg-dark: hsl(0deg 100% 0%);--contrast-dark: hsl(0deg 82% 38%);--contrast-2-dark: hsl(0deg 87% 31%)}
/*# sourceMappingURL=fun.css.map */
