.text-2xl{font-size:2rem;line-height:normal;text-align:center}@font-face{font-display:swap;font-family:clock-face;font-style:normal;src:url(/static/fonts/cursedtimerulil-aznm.woff2?v=0) format("woff2")}#uptime{font-style:normal}
/*# sourceMappingURL=uptime.css.map */
