{% extends "../../base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/quotes/quotes.css')}}">
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/quotes/info.css')}}">
{% end %}

{% block body %}
    <div class="info-page-0">
        <div class="info-page-data">{% block info_column %}{% end %}</div>
        <div {% if not wrong_quotes %}style="flex:min-content"{% end %}
            class="info-page-q-col">
            <span>
                 <a class="no-style-a button-container-a"
                    tooltip="Neues falsches Zitat hinzufügen"
                    href="{{fix_url('/zitate/erstellen', **create_kwargs)}}">
                     <img class="button-img" alt="Zitat hinzufügen."
                         src="{{fix_static('img/plus.svg')}}">
                 </a>
                {% if wrong_quotes %}<b>Falsche Zitate:</b> {% end %}
            </span>

            {% if wrong_quotes %}
                <ol>
                    {% for wq in wrong_quotes %}
                        <li>
                            <a href="{{fix_url(f'/zitate/{wq.get_id_as_str()}')}}"
                               class="wrong-quote-a">{{str(wq)}}</a>
                            (Rating: <b class="rating {% if wq.rating > 0 %}positive{% elif wq.rating < 0 %}negative{% end %}"
                                            >{{wq.rating}}</b>)
                        </li>
                    {% end %}
                </ol>
            {% else %}
                <span>Keine falschen Zitate gefunden.</span>
            {% end %}
        </div>
    </div>
    <span data-elastic-exclude id="disclaimer">
        <div>
            Die auf dieser Seite präsentierten Informationen sind nicht notwendigerweise
            korrekt. Wenn Du einen Fehler gefunden hast, kontaktiere uns bitte. Danke :)
        </div>
        {% if settings.get("CONTACT_USE_FORM") %}
            <a href="{{fix_url('/kontakt', subject=f'Fehler bei {type} {id}', message=f'{text}\n({canonical_url})')}}"
               tooltip="Melde Fehler"
               tooltip-position="left"
               rel="nofollow"
               class="button-container-a"
               id="report">
                <img class="report-img" alt="Fehler melden" src="{{fix_static('img/report.svg')}}">
            </a>
        {% end %}
    </span>
{% end %}
