.button,p{max-width:50rem;font-size:1.1rem}.button{width:calc(100% - 2.5rem);max-width:47.5rem;padding:.875rem 1.25rem;border:none;border-radius:.25rem;margin:.5rem 0;color:var(--text-color);cursor:pointer;text-align:center}.button:hover{filter:contrast(120%) brightness(120%);transform:scale(1.01)}.leave{background-color:var(--main)!important}.stay{background-color:var(--contrast)!important}
/*# sourceMappingURL=redirect.css.map */
