:root{--main-light: hsl(279deg 100% 57%);--main-2-light: hsl(279deg 90% 69%);--text-color-light: hsl(0deg 0% 0%);--secondary-text-color-light: hsl(0deg 0% 15%);--bg-light: hsl(180deg 30% 90%);--secondary-bg-light: hsl(180deg 10% 75%);--tertiary-bg-light: hsl(180deg 10% 60%);--contrast-light: hsl(0deg 100% 33%);--contrast-2-light: hsl(5deg 90% 22%);--main-dark: hsl(279deg 90% 69%);--main-2-dark: hsl(279deg 100% 57%);--text-color-dark: hsl(0deg 0% 90%);--secondary-text-color-dark: hsl(0deg 0% 70%);--bg-dark: hsl(0deg 0% 0%);--secondary-bg-dark: hsl(0deg 0% 9%);--tertiary-bg-dark: hsl(0deg 0% 18%);--contrast-dark: hsl(0deg 90% 48%);--contrast-2-dark: hsl(0deg 90% 62%)}
/*# sourceMappingURL=purple.css.map */
