import{jsx as v}from"/static/js/utils/utils.js?v=c676bb123f83b0ed";// @license magnet:?xt=urn:btih:d3d9a9a6595521f9666a5e94cc830dab83b65699&dn=expat.txt MIT
var l=document.getElementById("snow"),h=200;function $(n){n?l.style.display="block":l.style.display="none"}function b(n){document.querySelector("head").append(v("style",{children:n}))}function p(n=100){return Math.floor(Math.random()*n)+1}function y(n,t){return n=Math.ceil(n),t=Math.floor(t),Math.floor(Math.random()*(t-n+1))+n}function w(n,t){return Math.random()*(t-n)+n}function g(n=200){let t=[];for(let o=1;o<=n;o++){let e=Math.random()*100,r=Math.random()*10,i=e+r,d=e+r/2,s=w(.3,.8),c=s*100,a=Math.random(),u=y(10,30),m=p(30)*-1,f=Math.random();t.push("#snow p:nth-child(".concat(o,"){opacity:").concat(f,";transform:translate(").concat(e,"vw,-10px) scale(").concat(a,");animation:fall-").concat(o," ").concat(u,"s ").concat(m,"s linear infinite}@keyframes fall-").concat(o,"{").concat(s*100,"%{transform:translate(").concat(i,"vw,").concat(c,"vh) scale(").concat(a,")}to{transform:translate(").concat(d,"vw,100vh) scale(").concat(a,")}}"))}b(t)}function M(){g(h)}export{M as createSnow,$ as showSnow};
// @license-end
//# sourceMappingURL=snow.js.map
