.form,form{width:min(50rem,calc(100% - 2rem))}textarea,select,input[type=email],input[type=number],input[type=range],input[type=search],input[type=text]{display:block;width:100%;box-sizing:border-box;padding:.75rem 1.25rem;border:.125rem solid var(--main-2);border-radius:.625rem;margin:.5rem 0;background:var(--bg)}input[type=range]{padding:initial;border:initial}input[type=submit],button[type=submit],.form-button{width:100%;padding:.875rem 1.25rem;border:none;border-radius:.25rem;margin:.5rem 0;background-color:var(--secondary-bg);cursor:pointer}input[type=submit]:hover,button[type=submit]:hover,.form-button:hover{background-color:var(--main)}.form-element{display:block;width:100%;margin-bottom:.625rem}
/*# sourceMappingURL=form.css.map */
