{% extends "../base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/table.css')}}">
{% end %}

{% block body %}
    {% for service in services %}
        <h2>{{service.title}}</h2>
        <div style="max-width: 50rem">{{service.text}}</div>
        {% if service.infos %}
            <table class="table">
                <tbody>
                    {% for key, value in service.infos.items() %}
                        <tr>
                            <td>{{key}}</td>
                            {% if value.startswith('http') and '://' in value %}
                                <td><a href="{{fix_url(value)}}">{{value}}</a></td>
                            {% else %}
                                <td>{{value}}</td>
                            {% end %}
                        </tr>
                    {% end %}
                </tbody>
            </table>
        {% end %}
    {% end %}
{% end %}
