.comic-header{color:var(--text-color);text-decoration:none}input{width:100%;font-size:1.44rem}.normal-img{display:block;max-width:100%;height:auto;max-height:90%;margin:0 auto}#current-img{width:100%;max-height:80%}#old-comics-list{max-width:58%;padding:0;margin:0 auto;list-style-type:none}.popup-container{position:fixed;top:50%;left:50%;display:block;overflow:auto;width:100%;max-height:80%;margin:0 auto;transform:translate(-50%,-50%)}.popup-img{max-width:100%;max-height:100%}.close-button{position:absolute;top:.1rem;right:.1rem;width:2.625rem}
/*# sourceMappingURL=kangaroo_comics.css.map */
