{% extends "base.html" %}

{% block title %}{{status}} - {{reason}}{% end %}

{% block description %}{{description}}{% end %}

{% block head %}
    <meta name="robots" content="noindex">
{% end %}
{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/error.css')}}">
{% end %}

{% block body %}
    {% if is_traceback %}
        <pre>{{reason}}</pre>
    {% else %}
        <div id="container">
            <div id="status">{{status}}</div>
            <div id="reason" class="mono-font">{{reason}}</div>
        </div>
    {% end %}
{% end %}
