{% extends "../base.html" %}

{% block head %}
    <meta name="robots" content="noindex">
{% end %}
{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/redirect.css')}}">
{% end %}

{% block body %}
    <p>
        Willst du wirklich <strong>{{redirect_url.split("?")[0]}}</strong> öffnen? Für die Inhalte auf
        fremden Seiten können wir keine Garantie bieten. Weitergehen ist auf eigene Gefahr.
        {# {% if discord and not testing %}
            <br>
            Wenn du Discord nicht verwenden möchtest, kannst du auch Matrix nutzen,
            da bei uns alle Nachrichten gebridget werden. Dies kannst du zum Beispiel
            über <a no-dynload href="{{fix_url('/chat')}}">chat.asozial.org</a> tun.
        {% end %} #}
    </p>
    <a class="leave button no-style-a" no-dynload
       referrerpolicy="{{'strict-origin-when-cross-origin' if send_referrer else 'same-origin'}}"
       href="{{redirect_url}}">Ja, ich bin mir sicher</a>
    {% if from_url %}
        <a class="stay button no-style-a"
           href="{{from_url}}"
        >{{globals().get("back_button_text") or "Nein, ich möchte wieder zurück" }}</a>
    {% end %}
{% end %}
