{% extends "../base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" type="text/css" href="{{fix_static('css/form.css')}}">
{% end %}

{% block body %}
    <p>Hallo, {{name}}!</p>

    <form method="get">
        <label>
            Name:
            <input type="text" name="name" value="{{name}}">
        </label>

        <button type="submit">Eingabe</button>
    </form>
{% end %}
