input[type=radio]:checked+label[for]{color:var(--contrast-2)}input[type=radio]{display:inline-block;margin-right:.2rem}.label-input-group{display:block;max-width:calc(100% - .4rem);padding:.25rem;border-radius:.1rem}fieldset{padding:.3rem;border-color:var(--main);border-radius:.5rem}fieldset .label-input-group:nth-of-type(odd){background-color:var(--bg)}fieldset .label-input-group:nth-of-type(2n){background-color:var(--secondary-bg)}
/*# sourceMappingURL=generator.css.map */
