{% extends "../base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/form.css')}}">
{% end %}

{% block body %}
    <form method="get">
        {% raw form_appendix %}
        <label>Text zum Tröten
            <textarea name="text">{{text}}</textarea>
        </label>
        <label>Mastodon-Instanz
            <input name="instance" type="text" value="{{instance}}">
        </label>
        <label tooltip="Speichere die eingegebene Instanz in einem Cookie"
               tooltip-position="right">Speichere Instanz
            <input type="checkbox"
                   {% if save %} checked {% end %}
                   name="save"
                   value="sure">
        </label>
        <button tooltip="Teile auf Mastodon" type="submit">Tröt!</button>
    </form>
{% end %}
