{% extends "info.html" %}

{% block info_column %}
    <b>Autor: {{author}}</b>
    {% if author.info is not None %}
        <div>
            <p id="author-info">{{ author.info[1] }}</p><br>
            Quelle: <a href="{{fix_url(author.info[0])}}">Wikipedia</a>
            (Stand: {{author.info[2].date()}})
        </div>
    {% end %}
{% end %}
