{% extends "../../base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/form.css')}}">
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/quotes/generator.css')}}">
{% end %}

{% block body %}
    <form method="get" action="/zitate">
        {% raw form_appendix %}
        <h2>Wähle ein Zitat aus</h2>
        <fieldset id="quotes">
            {% for quote in quotes %}
                <div class="label-input-group"
                     tooltip="Autor*in: {{quote.author.name}}"
                     tooltip-position="bottom">
                    <input type="radio" value="{{quote.id}}"
                           id="q{{quote.id}}" name="quote">
                    <label for="q{{quote.id}}">{{quote.quote}}</label>
                </div>
            {% end %}
        </fieldset>
        <h2>Wähle einen Autor aus</h2>
        <fieldset>
            {% for author in authors %}
                <div class="label-input-group">
                    <input type="radio" value="{{author.id}}"
                           id="a{{author.id}}" name="author">
                    <label for="a{{author.id}}">{{author.name}}</label>
                </div>
            {% end %}
        </fieldset>
        <button type="submit">Einreichen</button>
    </form>
{% end %}
