{% extends "../base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" type="text/css"
          href="{{fix_static('css/form.css')}}">
{% end %}

{% block body %}
    <form method="post">
        <label>Name<input type="text" name="name"></label>

        <label>E-Mail-Adresse<input type="email" name="addresse" autocomplete="email"></label>

        <label>
            Betreff
            {% if subject %}
                <input type="text" name="subjekt" value="{{subject}}">
            {% else %}
                <input type="text" name="subjekt">
            {% end %}
        </label>

        {% comment 🍯 %}
        <label class="hidden">
            It's message time!
            <textarea id="message-input" name="message"></textarea>
        </label>

        <label>
            Nachricht (erforderlich)
            <textarea name="nachricht" required>{{message}}</textarea>
        </label>

        <button type="submit">Absenden</button>
    </form>
{% end %}
