:root{--main-light: hsl(100deg 100% 30%);--main-2-light: hsl(100deg 100% 40%);--text-color-light: hsl(0deg 0% 0%);--secondary-text-color-light: hsl(0deg 0% 15%);--bg-light: hsl(264deg 29% 90%);--secondary-bg-light: hsl(0deg 20% 86%);--tertiary-bg-light: hsl(180deg 10% 60%);--contrast-light: hsl(0deg 100% 33%);--contrast-2-light: hsl(4deg 87% 30%);--main-dark: hsl(100deg 100% 40%);--main-2-dark: hsl(100deg 100% 30%);--text-color-dark: hsl(100deg 0% 90%);--secondary-text-color-dark: hsl(250deg 0% 60%);--bg-dark: hsl(0deg 0% 0%);--secondary-bg-dark: hsl(0deg 0% 9%);--tertiary-bg-dark: hsl(0deg 0% 18%);--contrast-dark: hsl(0deg 90% 48%);--contrast-2-dark: hsl(0deg 90% 62%)}
/*# sourceMappingURL=green.css.map */
