{% extends "../base.html" %}

{% block description %}{{ version }}{% end %}

{% block body %}
{% whitespace single %}

<pre style="{% if full %}max-width: 100%; overflow: scroll; scrollbar-width: none{% else %}font-size: 1rem{% end %}">
Version: {{ version }}
    {% if no_3rd_party %}
<a href="{{fix_url(f'{GH_PAGES_URL}/coverage')}}"
   no-dynload>Code-Coverage</a>
    {% else %}
Code-Coverage: <a href="{{fix_url(f'{GH_PAGES_URL}/coverage')}}"
                  class="no-style-a"><img src="{{GH_PAGES_URL}}/coverage/badge.svg"
                                             no-dynload crossorigin></a>
    {% end %}
<a href="{{fix_url(f'{GH_PAGES_URL}/hashes.txt')}}"
   no-dynload>Hashes der neuesten Version</a><br>

Hash der Datei-Hashes:
{{ hash_of_file_hashes }}<br>
{% if full %}
Datei-Hashes:
{{ file_hashes }}{% end %}</pre>

{% end %}
