{% extends "base.html" %}

{% block stylesheets %}
    <link rel="stylesheet" href="{{fix_static('css/ansi2html.css')}}">
{% end %}

{% block body %}
    {% whitespace all %}

    {% if isinstance(ansi, list) %}
<div>
        {% for _ in ansi %}
<div class="pre">{% raw ansi2html(_) %}</div>
        {% end %}
</div>
    {% else %}
<div class="pre">{% raw ansi2html(ansi) %}</div>
    {% end %}

    {% whitespace oneline %}
    <br>
    <div style="display: block">
        Powered by: <a href="{{fix_url(powered_by)}}">{{powered_by_name}}</a>
    </div>
{% end %}
