[tooltip]{position:relative}[tooltip]:before,[tooltip]:after{position:absolute;z-index:1000000000;display:none;font-size:1rem;opacity:0;pointer-events:none;visibility:hidden}[tooltip]:after{width:max-content;max-width:max(100%,8.8rem);padding:.4rem;border-width:thin;border-style:solid;border-color:var(--tp-border-color);border-radius:.4rem;background-color:var(--tp-bg-color);color:var(--tp-text-color);content:attr(tooltip);font-family:ui-sans-serif,system-ui,sans-serif;-webkit-hyphens:initial;hyphens:initial;letter-spacing:normal;opacity:1;text-align:center;white-space:pre-line}[tooltip]:before{border-width:.3125rem;border-style:solid;border-color:transparent;content:" "}[tooltip]:hover:after,[tooltip]:hover:before,[tooltip]:focus:after,[tooltip]:focus:before,[tooltip].show-tooltip:after,[tooltip].show-tooltip:before{display:inherit;opacity:1;transition:all .05s ease .1s;visibility:visible}[tooltip][tooltip-position=top]:after,[tooltip][tooltip-position=top]:before,[tooltip]:not([tooltip-position]):after,[tooltip]:not([tooltip-position]):before,[tooltip][tooltip-position=bottom]:after,[tooltip][tooltip-position=bottom]:before{left:50%;transform:translate(-50%)}[tooltip][tooltip-position=left]:after,[tooltip][tooltip-position=left]:before,[tooltip][tooltip-position=right]:after,[tooltip][tooltip-position=right]:before{top:50%;transform:translateY(-50%)}[tooltip][tooltip-position=top]:after,[tooltip]:not([tooltip-position]):after{bottom:calc(100% - var(--tp-offset))}[tooltip][tooltip-position=bottom]:after{top:calc(100% - var(--tp-offset))}[tooltip][tooltip-position=left]:after{right:calc(100% - var(--tp-offset))}[tooltip][tooltip-position=right]:after{left:calc(100% - var(--tp-offset))}[tooltip][tooltip-position=top]:before,[tooltip]:not([tooltip-position]):before{top:var(--tp-offset);border-top-color:var(--tp-border-color)}[tooltip][tooltip-position=bottom]:before{bottom:var(--tp-offset);border-bottom-color:var(--tp-border-color)}[tooltip][tooltip-position=left]:before{left:var(--tp-offset);border-left-color:var(--tp-border-color)}[tooltip][tooltip-position=right]:before{right:var(--tp-offset);border-right-color:var(--tp-border-color)}
/*# sourceMappingURL=tooltip.css.map */
