{% extends "../../base.html" %}

{% block head %}
    <link rel="me" href="https://fosspri.de/@asozialer_bot">
    {% include "./link_rss.html" %}
{% end %}

{% block body %}
    <p class="center-text">
        Hier gibt es ganz viele bewertbare falsch zugeordnete Zitate.
        Die Idee hinter diesen stammt aus dem
        <a class="no-style-a" href="{{fix_url('https://marcuwekling.reimkultur-shop.de/marc-uwe-kling-das-kaenguru-manifest-live-und-ungekuerzt-download-album-6400139.html')}}"
        >Känguru-Manifest</a>, in welchem
        {% if kangaroo_url %}
            <a class="no-style-a" href="{{kangaroo_url}}">das Känguru</a>
        {% else %}
            das Känguru
        {% end %}
        sich diese ausdenkt und
        {% if muk_url %}
            <a class="no-style-a" href="{{muk_url}}">Marc-Uwe Kling</a>
        {% else %}
            Marc-Uwe Kling
        {% end %}
        davon berichtet.
        Falls Du Dich nach dem Sinn hinter diesen fragst, dann lass mich
        {% if one_stone_url %}
            <a class="no-style-a"
               href="{{one_stone_url}}">Albert Einstein</a>
        {% else %}
            Albert Einstein
        {% end %}
        zitieren:
    </p>
    <blockquote>
        Falsch zugeordnete Zitate sind wie Physik, unnötig aber einfach witzig.
    </blockquote>
    <p class="center-text">
        Teilweise kommen da echt
        <a href="{{fix_url('/vertauschte-woerter/?text=relevant')}}"
           class="no-style-a">relevante</a> Zitate bei herum,
        die Du dann mit Deinen Freunden Teilen kannst.
        Schau Dich gerne mal um. Viel Spaß!
    </p>
    {% if quote_of_the_day %}
        <h2>Das Zitat des Tages
            <a href="{{fix_url('/zitat-des-tages/feed')}}"
               class="no-style-a" no-dynload
            ><img src="{{fix_static('img/rss.svg')}}"
                  alt="RSS-Feed" style="width:1rem"></a></h2>
        <a class="no-style-a center-text"
           href="{{fix_url(quote_of_the_day.get_quote_url())}}">
            {{quote_of_the_day.wrong_quote.quote}}<br>
            - {{quote_of_the_day.wrong_quote.author}}
        </a>

    {% end %}
    <h2>Links zu Zitaten</h2>
    <ul>
        <li>
            <a href="{{funny_quote_url}}">Witzige Zitate</a>
        </li>
        <li>
            <a href="{{random_quote_url}}">Zufällige Zitate</a>
        </li>
    </ul>
    <h2>Zitate beeinflussen</h2>
    <ul>
        <li>
            <a href="{{fix_url('/zitate/erstellen')}}">Zitate erstellen</a>
        </li>
        <li>
            <a href="{{fix_url('/zitate/generator')}}">Zitate-Generator</a>
        </li>
        {% if settings.get("CONTACT_USE_FORM") %}
            <li>
                <a href="{{fix_url('/kontakt', subject='Falsch zugeordnete Zitate')}}"
                   rel="nofollow"
                >Über Zitate beschweren</a>
            </li>
        {% end %}
    </ul>
    <br>
{% end %}
