#container *{color:var(--main-2);text-align:center}#status{font-size:max(4rem,min(16rem,32vw))}#reason{font-size:max(1rem,min(4rem,8vw));-webkit-hyphens:none;hyphens:none}
/*# sourceMappingURL=error.css.map */
