.rating{display:inline;color:var(--secondary-text-color)}.positive{color:green;color:light-dark(green,#0f0)}.negative{color:red}.wrong-quote-a{color:var(--text-color)!important}.info-page-0{display:flex;flex-direction:column;gap:2rem;place-content:center}@media only screen and (width >= 768px){.info-page-0{flex-direction:row}.info-page-0 div{flex:30%}}.info-page-data{display:flex;flex-direction:column;place-content:center;text-align:center}.info-page-q-col{display:flex;flex-direction:column;place-content:center}.info-page-q-col span{display:flex;flex-direction:row;gap:1ch;place-content:center}#disclaimer{display:flex;flex-direction:row;margin-top:2rem;gap:1rem}.button-container-a{--img-size: 1rem;display:inline-block;width:var(--img-size);height:var(--img-size)}#report{--img-size: 3.2rem}.button-container-a img{display:block;width:var(--img-size);height:var(--img-size)}#author-info{padding:.3125rem;border-right:.3125rem;background-color:var(--secondary-bg)}
/*# sourceMappingURL=info.css.map */
