.button-img{width:4.3rem;height:4.3rem;border-radius:50%}.button-img:hover{filter:drop-shadow(3px 3px 2px rgb(139 0 0 / 50%));transform:scale(1.08)}.report-img:hover{filter:drop-shadow(3px 3px 2px rgb(217 229 78 / 50%))}#quote-container{display:block;height:auto;padding:4rem;background:transparent;text-align:center}#top{background-color:transparent}.quote-text{font-size:1.9rem;line-height:1.8rem;text-align:left}.quote-author{margin-top:1rem;margin-bottom:1rem;font-size:1.4rem;text-align:right}#rating-container{align-items:start;font-size:1.4rem;text-align:left}#rating-container .flex{display:flex;flex-flow:row}#rating-text{display:block;margin-right:.3rem}.rating-img{display:inline-block}.nicht-witzig{bottom:.008rem}.nicht-witzig:after{content:url(/static/img/stamps/nichtwitzig.small.svg?v=252533840141546d)}.witzig:after{content:url(/static/img/stamps/witzig.small.svg?v=af22bf11feffb9bb)}.invisible{display:none;width:0;height:0;opacity:0}#bottom-buttons-container{display:flex;justify-content:left}.rating-form-el{display:inline-block;width:auto!important}#add-quote,#report{margin-right:.69rem}#buttons a{position:relative;display:inline-block;color:#8b0000}.download-img{transform:rotate(90deg)}.download-img:hover{filter:drop-shadow(3px -3px 2px rgb(139 0 0 / 50%));transform:scale(1.08) rotate(90deg)}#vote-container{display:flex;min-width:69%;align-items:stretch;place-content:stretch center}.vote{display:inline-block;width:auto!important;max-width:initial;height:3.75rem;flex-grow:1;padding:.625rem!important;margin:.3125rem}.vote:hover:not([voted]){background-color:var(--tertiary-bg)!important}.vote[voted]{filter:grayscale(69%)}.vote img{width:auto;max-height:2.5rem;background-color:transparent!important}@media (orientation: landscape){#quote-container{position:relative;z-index:1;display:block;width:50rem;max-width:90%;min-height:23.75rem;padding:0 2rem 2rem;margin:1rem auto;background:transparent url(/static/img/kangarooface.svg?v=0) no-repeat right top .3125rem;text-align:center}.quote-text{width:32rem}.quote-author{margin-right:18rem}#buttons{position:absolute;right:0;bottom:0}#rating-container{position:absolute;bottom:1.25rem}}
/*# sourceMappingURL=quotes.css.map */
